
type tree =
  | E
  | N of tree * int * tree

(* une fonction de hachage (arbitraire) pour les arbres *)

let rec hash t = match t with
  | E -> 1
  | N (l, n, r) -> 19 * (19 * hash l + n) + hash r

(* calcul du nombre de noeuds N physiquement différents
   (donc, à un facteur près, la taille occupée par l'arbre en mémoire)

   un simple parcours qui ajoute tous les sous-arbres dans un ensemble *)
let mem_size t =
  let h = Hashset.create 5003 in
  let rec add t = match t with
    | E -> ()
    | N (l, _, r) -> Hashset.add hash (==) h t; add l; add r
  in
  add t;
  Hashset.size h

(* mais c'est un peu dommage de (re)parcourir entièrement un arbre déjà
   rencontré ; on peut donc procéder ainsi *)
let mem_size t =
  let h = Hashset.create 5003 in
  let rec add t = match t with
    | E -> ()
    | N (l, _, r) ->
      match Hashset.find hash (==) h t with
        | None -> Hashset.add hash (==) h t; add l; add r
        | Some _ -> ()
  in
  add t;
  Hashset.size h

(* un inconvénient, cependant : quand find renvoie None, on devra recalculer
   la valeur de hachage pour add. On pourrait améliorer cela avec une
   fonction add qui indique si l'ajout a effectivement été réalisé (avec
   un booléen par exemple). Ce n'est pas demandé ici. *)

(* partage maximal : étant donné un arbre, on renvoie un arbre
   structurellement identique, mais dans lequel tous les sous-arbres
   égaux sont uniques en mémoire *)

let share t =
  let h = Hashset.create 5003 in
  let rec make t = match t with
    | E -> E
    | N (l, n, r) as t ->
      match Hashset.find hash (=) h t with
        | None ->
          let t = N (make l, n, make r) in
          Hashset.add hash (=) h t;
          t
        | Some t ->
          t
  in
  make t

(* on peut éviter de calculer deux fois la clé de hachage (find + add)
   en utilisant une fonction find_or_add (non demandée) mais on construit
   alors inutilement le noeud N dans certains cas (et on fait des appels
   récursifs inutilement) *)
let share t =
  let h = Hashset.create 5003 in
  let rec make t = match t with
    | E -> E
    | N (l, n, r) ->
        Hashset.find_or_add hash (=) h (N (make l, n, make r))
  in
  make t

(* optimisation : si on fait le partage des sous-arbres en premier lieu,
   alors on peut utiliser une égalité en O(1) qui fait directement de
   l'égalité physique sur les sous-arbres *)

let equal t1 t2 = match t1, t2 with
  | E, E -> true
  | N (l1, n1, r1), N (l2, n2, r2) -> l1 == l2 && n1 = n2 && r1 == r2
  | E, N _ | N _, E -> false

let share t =
  let h = Hashset.create 5003 in
  let rec make t = match t with
    | E -> E
    | N (l, n, r) ->
      let t = N (make l, n, make r) in
      match Hashset.find hash equal h t with
        | None -> Hashset.add hash equal h t; t
        | Some t -> t
  in
  make t
